// C Source File
// Created 05/31/2003; 14:59:54

/* 
	Tally Counter v.1.00.00
	This program is used to simulate a mechanical counting device.
	Copyright (C) 2003 Travis Evans
	
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#define USE_TI89

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization

#define SAVE_SCREEN           // Save/Restore LCD Contents
#define SET_FILE_IN_USE_BIT   // Just to be safe

#include <tigcclib.h>         // Include All Header Files

#include "tlycnt10000.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Counter data
     This data will be saved in the executable and re-loaded on each program start
     c_mode = current counter mode; mainly determines # digits from 1 to 5.
     c_global = indicates whether separate tax, visual effects, and operation modes are to be used for
       each counter, or if the settings will be global to all counter modes. When global
       settings is turned on, the current counter's settings will be copied to another separate
       set of settings variables, and these will be used for all counter modes. Operation returns to
       using the arrays to store settings when global settings is turned off.
     c_count = actual counter value for each counter
     c_tax = tax value for each counter
     c_opMode = current operation mode for each counter (either +1 or -1 for add or subtract mode)
     c_effects = determines whether visual effects are enabled for each counter (0=no, 1=yes)
*/

static short 		c_mode = 3;
static BOOL		c_global = FALSE;
static signed long	c_count[] = 		{0, 0, 0, 0, 0};
static signed long 	c_tax[] =		{0, 0, 0, 0, 0};
static char	 	c_opMode[] =		{'+', '+', '+', '+', '+'};
static BOOL 		c_effects[] =		{FALSE, FALSE, FALSE, FALSE, FALSE};
static signed long	cG_tax =		0;			// Global settings
static BOOL		cG_effects =		FALSE;			//   stored here
static char		cG_opMode =		'+';			//

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* A few tiny functions that return certain option settings, taking into account
   current counter mode and whether Global Settings is turned on. This is mainly
   to save some bytes in the compiled object code, since these functions are
   called in many places. */

BOOL CurrEffectsMode(void) {
	return c_global ? cG_effects : c_effects[c_mode];
}

char CurrOpMode(void) {
	return c_global ? cG_opMode : c_opMode[c_mode];
}

long CurrTaxValue(void) {
	return c_global ? cG_tax : c_tax[c_mode];
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Returns the specified integer power of 10.
     Valid input values are positive integers >= 0. */

signed long PowOf10(short power) {
	signed long result = 1;
	short count = 1;
	for (count = 1; count <= power; count++) {
		result *= 10;
	}
	return result;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Checks validity of counter value (i.e., not over its range) and corrects the counter
   if necessary.
*/

signed long ValidateCounterVal(signed long val) {

	signed long maxLimit = PowOf10(c_mode+1) - 1;
	if (val > maxLimit)
		val %= (maxLimit + 1);
	while (val < 0)
		val += (maxLimit+1);
	return val;
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Draws a window with a single-line border and 2-pixel wide shadow on the screen
   Input the coordinates of the upper-left corner where the border is and the lower-right corner of
     where the border is. The shadow will extend past the right and lower edges of the
     specified window.
   The title parameter displays an optional title in the 4x6 font over the window. This can be omitted
     by passing a null string.
*/
void DrawWindow(short x1, short y1, short x2, short y2, const char *title) {
	
	// Save current font setting (so it isn't messed up when we return to calling routine)
	short prevFont = FontGetSys();
	
	// Border
	DrawLine(x1, y1, x2, y1, A_NORMAL);				// Top border
	DrawLine(x2, y1, x2, y2, A_NORMAL);				// Right border
	DrawLine(x2, y2, x1, y2, A_NORMAL);				// Bottom border
	DrawLine(x1, y2, x1, y1, A_NORMAL);				// Left border
	
	// Shadow
	DrawLine(x2+1, y1+2, x2+1, y2+2, A_NORMAL);			// Right shadow
	DrawLine(x2+2, y1+2, x2+2, y2+2, A_NORMAL);			//
	DrawLine(x1+2, y2+1, x2+2, y2+1, A_NORMAL);			// Bottom shadow
	DrawLine(x1+2, y2+2, x2+2, y2+2, A_NORMAL);			//
	
	// Draw title for window
	FontSetSys(F_4x6);
	DrawStr(x1, y1-5, title, A_REVERSE);
	
	// Restore font and return
	FontSetSys(prevFont);
	return;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Draw main interface of the program on the screen
*/
void DrawUI(void) {
	
	short xCoord, yCoord;
	ClrScr();
	DisplayOff();							//Blank screen to make interface look faster
									//  than it is for improved appearance
	FontSetSys(F_6x8);
	DrawLine(0, LCD_HEIGHT-7, LCD_WIDTH-1, LCD_HEIGHT-7, A_NORMAL);	//Leave room for status bar	

	// Draw background pattern
	for (yCoord=0; yCoord<=8; yCoord++) {
		DrawLine(0, yCoord, (LCD_WIDTH-1), yCoord, A_NORMAL);	//Black bar at top of screen where title goes
	} for (xCoord=0; xCoord <= (LCD_WIDTH-1); xCoord+=2) {
		DrawLine(xCoord, 9, xCoord, 6*8-1, A_SHADE_H);	//Background "gray" pattern
	}

	// Title bar	
	DrawStr(tCntrX(23), 0, "Tally Counter v.1.00.00", A_REVERSE);

	short hPos, vPos, rowCount, colCount;
	short index = 0;
	
	// Counter window
	DrawWindow(cntrHorzPos - 1, 3*8-1, cntrHorzPos + (6*(c_mode+3)), 4*8, "Counter");

	// Configuration window
	DrawWindow(configHorzPos - 1, 2*8-1, configHorzPos + 6*9, 6*8, "Configuration");
	
	// Help text
	DrawLine(0, 6*8-1, LCD_WIDTH, 6*8-1, A_NORMAL);
	FontSetSys(F_4x6);
	for (colCount = 0; colCount < keyHelpCols; colCount++) {
		hPos = keyHelpHStart[colCount];
		vPos = keyHelpVStart;
		for (rowCount = 0; rowCount < keyHelpRows; rowCount++) {
			DrawStr(hPos, vPos, keyHelpList[index], A_NORMAL);
			vPos += keyHelpVIncr;
			index++;
		}
	}
	
	// Done
	DisplayOn();
	
	return;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Update displayed data of the main interface of the program
*/

void UpdateUI (void) {

	FontSetSys(F_6x8);
	char buf[10];
	//Declaring this instead of writing the string literal actually appears
	//  to save a few bytes in the compiled program :-)
	static const char *threeSpaces = "   ";

	// Make sure counter value is within range before displaying
	c_count[c_mode] = ValidateCounterVal(c_count[c_mode]);

	// Counter digits
	sprintf(buf, " %0*ld ", c_mode+1, c_count[c_mode]);
	DrawStr(cntrHorzPos, 3*8, buf, A_REPLACE);
	
	// Configuration info
	sprintf(buf, " Oper: %c ", CurrOpMode());
	DrawStr(configHorzPos, 2*8, buf, A_REPLACE);
	sprintf(buf, " Mode: %d ", c_mode+1);
	DrawStr(configHorzPos, 3*8, buf, A_REPLACE);		
	sprintf(buf, " %s %s ",
		c_global ? "GLB" : threeSpaces,
		CurrEffectsMode() ? "EFX" : threeSpaces);
	DrawStr(configHorzPos, 4*8, buf, A_REPLACE);
	sprintf(buf, "Tax: %3ld%%", CurrTaxValue());
	DrawStr(configHorzPos, 5*8, buf, A_REPLACE);
	
	return;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Increments or decrements (according to current operation mode) the counter by
   the specified power of 10. The counter is adjusted even if the change would
   make it go over its range (i.e. 999 for three-digit counter mode). Therefore,
   ValidateCounter() should be called at some point to correct such a condition.
   
   The sense of the operation mode can be reversed if the 'reverse' parameter is
   set to TRUE. This is so that the normal sense of the operation mode can be
   reversed. For instance, pressing Diamond + F1...F5 simulates a temporary
   reversal of the current counter mode. This is done by setting the 'reverse'
   parameter to TRUE.
*/

void AdjCounter(short power, BOOL reverse) {
	if (CurrOpMode() == (reverse ? '-' : '+')) c_count[c_mode] += PowOf10(power);
	else c_count[c_mode] -= PowOf10(power);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* This routine produces an animated sequence of changing digits to change the counter
   value, only if visual effects is enabled. In producing visual effects, the routine
   takes the new value that the counter is to have. It uses the old counter value,
   animates the change to the new value on the screen, and then sets the counter value
   to the new number passed to it.
   
   If visual effects mode isn't turned on, the routine will not do any animation, it
   will just update the counter value on-screen and update the counter variable with
   the new value only. Therefore, this routine can be called unconditionally any time
   a change is made to the counter variable, and it will automatically conform to the
   current user configuration without checking it externally.
   
   The animation sequence gives the appearance that the digits are changing individually,
   one by one, until they match the new number.
*/

void EffectUpdateCounter(signed long newValue) {

	short digitNo, oldDigit, newDigit;
	signed long changeVal;
	
	// Keep both the old value and the new value in range to prevent
	//   endless loops
	c_count[c_mode] = ValidateCounterVal(c_count[c_mode]);
	newValue = ValidateCounterVal(newValue);
	
	if (CurrEffectsMode()) {

		// Visual effects routine
		// Can be terminated with a key press in case user doesn't want to wait for
		//   the animation or a bug occurs and an endless loop results
		while ((c_count[c_mode] != newValue) && (!kbhit())) {
			changeVal = 0;
			// Find out which digits to change
			for (digitNo = 1; digitNo <= c_mode+1; digitNo++) {
				
				oldDigit = (c_count[c_mode] / PowOf10(digitNo-1)) % 10;
				newDigit = (newValue / PowOf10(digitNo-1)) % 10;
				
				// Compare the digits from the old and new values and
				//   determine what value to add for this animation iteration.
				if (oldDigit < newDigit) {
					changeVal += PowOf10(digitNo-1);
				} else if (oldDigit > newDigit) {
					changeVal -= PowOf10(digitNo-1);
				}
			}
			
			// Display this "frame" of the animation
			c_count[c_mode] += changeVal;
			UpdateUI();
			
			// With a delay
			OSFreeTimer(USER_TIMER);
			OSRegisterTimer(USER_TIMER, 2);
			while (!OSTimerExpired(USER_TIMER));
		}

	}
	
	// Write the new value to the counter variable and update the screen
	c_count[c_mode] = newValue;
	UpdateUI();
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Uses a dialog box to get a counter number from the user.
     valSpec refers to the value to get (counter value or
     tax). The value will be set if the user enters a value
     and confirms the dialog box.
*/
     
void GetNumber(short valSpec) {

	HANDLE dlgHandle = DialogNewSimple(125, 40);
	char inputBuf[7];
	short result = 1;
	
	// Set up inputBuf to display the current value
	if (valSpec == COUNTER) {
		sprintf(inputBuf, "%.5ld", c_count[c_mode]);
	} else if (valSpec == TAX) {
		sprintf(inputBuf, "%.3ld", CurrTaxValue());
	}
	
	if (dlgHandle == H_NULL) {
		DlgMessage("Enter value", "Insufficient RAM to display dialog box", BT_OK, BT_NONE);
	} else {
		// Dialog box procedure
		short maxChars = (valSpec == COUNTER ? c_mode+1 : 3);
		DialogAddTitle(dlgHandle, titleStr[valSpec], BT_OK, BT_CANCEL);
		DialogAddRequest(dlgHandle, 5, 15, promptStr[valSpec], 0, maxChars, 6);
		result = DialogDo(dlgHandle, CENTER, CENTER, inputBuf, NULL);
	
		HeapFree(dlgHandle);		
		if (result == KEY_ENTER) {
			long value = atol(inputBuf);
			if (valSpec == COUNTER) c_count[c_mode] = value;
			if (valSpec == TAX) {
				if (c_global) cG_tax = value;
				else c_tax[c_mode] = value;
			}
		}
	}
	return;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Displays the main configuration menu that allows user to adjust program options.
*/

void ConfigMenu(void) {
	
	BOOL menuDone = FALSE;
	while (!menuDone) {

		// Dynamic menu setup
		short count;
		char globSet[] = " Global settings";				// Checkmark for global settings option
		if (c_global) globSet[0] = 15;					//
		char effModeSet[] = " Visual effects";				// Checkmark for global settings option
		if (CurrEffectsMode()) effModeSet[0] = 15;			//

		// Indicator for current counter mode, and for effects settings of counters
		// Dynamically generates the menu items for the counters
		char modeSet[5][9];
		const char *counterStr = "  -digit";
		short elm;
		for (count = 0; count <= 4; count++) {
			for (elm = 0; elm <= 8; elm++) {
				modeSet[count][elm] = counterStr[elm];
			}
			if (c_mode == count) modeSet[count][0] = 127;
			modeSet[count][1] = '1' + count;
		}
	
		// Configuration pop-up menu
		HANDLE menuHandle = PopupNew("Configuration", 0);
		if (menuHandle == H_NULL) {
			
			DlgMessage("Configuration menu",
				"Insufficient free RAM to display menu.",BT_OK,BT_NONE);
			return;
			
		} else {
			
			// Define popup menu
			PopupAddText(menuHandle, -1, globSet, 1);
			PopupAddText(menuHandle, -1, effModeSet, 2);
			PopupAddText(menuHandle, 0, " Counter mode", 3);
			PopupAddText(menuHandle, -1, " Counter set\xA0", 4);
			PopupAddText(menuHandle, -1, " Tax set\xA0", 5);
			PopupAddText(menuHandle, -1, " Reset all\xA0", 6);
			PopupAddText(menuHandle, -1, " Done", 99);
			// Mode settings loop
			short idCount = 7; short unsigned arrayPtr;
			for (arrayPtr = 0; arrayPtr <= 4; arrayPtr++) {
				PopupAddText(menuHandle, 3, modeSet[arrayPtr], idCount);
				idCount++;
			}
			
			short result = PopupDo(menuHandle, CENTER, CENTER, 0);
			
			switch (result) {
				case 1: c_global = !c_global; break;
				case 2:
					if (c_global) cG_effects = !cG_effects;
					else c_effects[c_mode] = !c_effects[c_mode];
					break;
				case 4:
					GetNumber(COUNTER);
					break;
				case 5:
					GetNumber(TAX);
					break;
				case 6:
					if (DlgMessage("Reset all",
					"This will reset ALL counters AND their settings. Are you sure?",
					BT_YES, BT_NO)==KEY_ENTER) {
						// Returning settings to defaults
						c_global = FALSE;
						short counter;
						for (counter = 0; counter <= 4; counter++) {
							c_count[counter] = 0;
							c_tax[counter] = 0;
							c_opMode[counter] = '+';
							c_effects[counter] = FALSE;
						}
						cG_tax = 0; cG_effects = FALSE; cG_opMode = '+';
					}
					break;
				case 7 ... 11:
					// Change counter mode
					c_mode = result - 7;
					DrawUI(); UpdateUI();
					break;
				case 0: case 99:
					// Done; exit the routine
					menuDone = TRUE;
			}
			HeapFree(menuHandle);
		}
	}
	return;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Performs an action based on a key press. Returns true if ESC or HOME is pressed,
   thus exiting the program.
*/

BOOL ProcessKey(short key) {
	
	switch (key) {
		case KEY_F1 ... KEY_F5:
			// Increment digits
			AdjCounter(KEY_F5 - key, FALSE);
		break;
		case KEY_DF1 ... KEY_DF5:
			// Decrement digits
			AdjCounter(KEY_DF5 - key, TRUE);
		break;
		case KEY_ESC: case KEY_HOME: case KEY_QUIT:
			// Exit program
			return TRUE;
		case KEY_OFF: case KEY_OFF2:
			off();
		break;
		case KEY_MODE:
			// Configuration menu
			ConfigMenu();
		break;
		case KEY_BACKSPACE:
			// Cycle through counter modes
			c_mode = ((c_mode + 1) % (sizeof(c_count)/sizeof(long)));
			DrawUI();				// Redraw screen for new counter mode
		break;
		case KEY_CLEAR:
			// Reset counter
			EffectUpdateCounter(0);
		break;
		case KEY_POWER:
			// Toggle visual effects
			if (c_global) cG_effects = ! cG_effects;
			else c_effects[c_mode] = !c_effects[c_mode];
		break;
		case KEY_STO:
			// Store counter value directly
			GetNumber(COUNTER);
		break;
		case KEY_EE:
			// Store tax value directly
			GetNumber(TAX);
		break;
		case KEY_SIGN:
			// Toggle operation mode (add, subtract)
			if (c_global)
				cG_opMode = (cG_opMode == '+' ? '-' : '+');
			else
				c_opMode[c_mode] = (c_opMode[c_mode] == '+' ? '-' : '+');
		break;
		case '/':
			// Divide by 2
			EffectUpdateCounter(c_count[c_mode] / 2);
		break;
		case '*':
			// Multiply by 2
			EffectUpdateCounter(c_count[c_mode] * 2);
		break;
		case '-':
			// Subtract tax
			if (CurrTaxValue()  != 0)
				EffectUpdateCounter( c_count[c_mode] - ( ( c_count[c_mode] * CurrTaxValue() ) / 100 ) );
		break;
		case '+':
			// Add tax
			if (CurrTaxValue() != 0)
				EffectUpdateCounter( c_count[c_mode] + ( ( c_count[c_mode] * CurrTaxValue() ) / 100 ) );
	}
	
	return FALSE;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//						***** MAIN *****
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Main Function
void _main(void)
{
	BOOL programDone = FALSE;
	
	// Set up screen
	DrawUI();
	
	// Main loop
	while (!programDone) {
		UpdateUI();
		programDone = ProcessKey(GKeyIn(NULL, 0));		// GKeyIn is more battery-efficient than ngetchx()
	}
}
